package main

/*
#include <stddef.h>
#include <stdint.h>

// We need this pointer to function machinery, because CGo tries to link the C
// part to compile successfully. This trick comes from pion examples.
// As an alternative, we could mess with the linker flags:
// https://github.com/golang/go/issues/25832

typedef uint8_t Uuid[16];
typedef struct CallerData *CallerDataPtr;
typedef void (*EventFunc)(Uuid, CallerDataPtr ctx);
typedef void (*MessageFunc)(Uuid, const uint8_t *data, size_t length, CallerDataPtr ctx);
typedef struct Callbacks {
	EventFunc hello;
	MessageFunc message;
	EventFunc bye;
} Callbacks;

inline void CallEvent(EventFunc f, Uuid u, CallerDataPtr ctx)
{
	f(u, ctx);
}
inline void CallMessage(MessageFunc f, Uuid u, uint8_t *data, size_t length, CallerDataPtr ctx)
{
	f(u, data, length, ctx);
}
*/
import "C"

import (
	"context"
	"encoding/json"
	"fmt"
	"time"
	"unsafe"

	"net/http"

	"github.com/google/uuid"
	"github.com/pion/webrtc/v3"
)

type Callbacks = C.Callbacks
type CallerData = C.CallerDataPtr

type Peer struct {
	conn    *webrtc.PeerConnection
	channel *webrtc.DataChannel
}

type PeerAddedEvent struct {
	id      uuid.UUID
	conn    *webrtc.PeerConnection
	channel *webrtc.DataChannel
}

type PeerSendEvent struct {
	id   uuid.UUID
	data []byte
}

var peerAdded chan PeerAddedEvent
var peerLeft chan uuid.UUID
var peerSend chan PeerSendEvent
var peerBroadcast chan []byte
var quit chan int
var callbacks Callbacks
var callerData CallerData

const ChannelName = "game"

func HttpHandler(w http.ResponseWriter, req *http.Request) {
	w.Header().Set("Access-Control-Allow-Origin", "*")
	w.Header().Set("Access-Control-Allow-Headers", "Content-Type")

	if req.Method == "OPTIONS" {
		return
	} else if req.Method != "POST" {
		http.Error(w, "Method not allowed", http.StatusMethodNotAllowed)
		return
	}

	config := webrtc.Configuration{
		ICEServers: []webrtc.ICEServer{
			{
				URLs: []string{"stun:stun.l.google.com:19302"},
			},
		},
	}

	myId, err := uuid.NewRandom()
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}
	cId := (*C.uchar)(unsafe.Pointer(&myId[0]))

	peerConnection, err := webrtc.NewPeerConnection(config)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	gotChannel := make(chan int)

	peerConnection.OnConnectionStateChange(func(s webrtc.PeerConnectionState) {
		if s == webrtc.PeerConnectionStateDisconnected {
			C.CallEvent(callbacks.bye, cId, callerData)
			peerLeft <- myId
		}
	})

	peerConnection.OnDataChannel(func(d *webrtc.DataChannel) {
		if d.Label() != ChannelName {
			return
		}
		d.OnOpen(func() {
			gotChannel <- 1
			C.CallEvent(callbacks.hello, cId, callerData)
			peerAdded <- PeerAddedEvent{
				id:      myId,
				conn:    peerConnection,
				channel: d,
			}
		})
		d.OnMessage(func(msg webrtc.DataChannelMessage) {
			data := (*C.uchar)(unsafe.Pointer(&msg.Data[0]))
			C.CallMessage(callbacks.message, cId, data, C.ulong(len(msg.Data)),
				callerData)
		})
	})

	offer := webrtc.SessionDescription{}
	err = json.NewDecoder(req.Body).Decode(&offer)
	if err != nil {
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}

	err = peerConnection.SetRemoteDescription(offer)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	answer, err := peerConnection.CreateAnswer(nil)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	err = peerConnection.SetLocalDescription(answer)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	gatherComplete := webrtc.GatheringCompletePromise(peerConnection)
	<-gatherComplete

	res, err := json.Marshal(*peerConnection.LocalDescription())
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}

	go func() {
		select {
		case <-time.After(30 * time.Second):
			peerConnection.Close()
		case <-gotChannel:
			return
		}
	}()

	w.Header().Set("Content-Type", "application/json")
	fmt.Fprintln(w, string(res))
}

//export WebrtcServerRun
func WebrtcServerRun(addr *C.char, theCallbacks Callbacks, theCallerData CallerData) {
	peerAdded = make(chan PeerAddedEvent)
	peerLeft = make(chan uuid.UUID)
	peerSend = make(chan PeerSendEvent)
	peerBroadcast = make(chan []byte)
	quit = make(chan int)
	callbacks = theCallbacks
	callerData = theCallerData

	peers := make(map[uuid.UUID]Peer)

	srv := &http.Server{Addr: C.GoString(addr)}
	http.HandleFunc("/", HttpHandler)
	go func() {
		srv.ListenAndServe()
	}()

	run := true
	for run {
		select {
		case added := <-peerAdded:
			peers[added.id] = Peer{
				conn:    added.conn,
				channel: added.channel,
			}
		case left := <-peerLeft:
			delete(peers, left)
		case send := <-peerSend:
			if peer, ok := peers[send.id]; ok {
				peer.channel.Send(send.data)
			}
		case buffer := <-peerBroadcast:
			for _, peer := range peers {
				peer.channel.Send(buffer)
			}
		case <-quit:
			run = false
		}
	}

	fmt.Println("Time to exit the WebRTC server!")
	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()
	if err := srv.Shutdown(ctx); err != nil {
		fmt.Printf("Could not stop the HTTP server: %s\n", err.Error())
	}

	for _, peer := range peers {
		if err := peer.conn.Close(); err != nil {
			fmt.Printf("Could not close a peer: %s\n", err.Error())
		}
	}
}

//export WebrtcServerStop
func WebrtcServerStop() {
	quit <- 0
}

//export WebrtcSend
func WebrtcSend(id *C.uchar, message *C.uchar, len C.size_t) {
	uuidSlice := unsafe.Slice((*byte)(unsafe.Pointer(id)), 16)
	uuidObj, _ := uuid.FromBytes(uuidSlice)
	data := unsafe.Slice((*byte)(unsafe.Pointer(message)), len)
	buffer := make([]byte, len)
	copy(buffer, data)
	peerSend <- PeerSendEvent{
		id:   uuidObj,
		data: buffer,
	}
}

//export WebrtcBroadcast
func WebrtcBroadcast(message *C.uchar, len C.size_t) {
	data := unsafe.Slice((*byte)(unsafe.Pointer(message)), len)
	buffer := make([]byte, len)
	copy(buffer, data)
	peerBroadcast <- buffer
}

func main() {}
